#include <avr/wdt.h>

#include "Arduino.h"

extern volatile float angle;  //从calcAngle文件引入angle变量
extern volatile unsigned char mpuToken;
void setup() {
    //注意HC-06 / HC-01 默认的波特率时9600
    //如果使用蓝牙遥控,切记要修改此处代码
    Serial.begin(115200);

    encoderSetup();
    pidSetup();
    mpuSetup();
    sei();

    //开启看门狗,如果1秒钟内没有执行wdt_rest,芯片将会复位
    wdt_enable(WDTO_1S);
    wdt_reset();
    pinMode(13, OUTPUT);
    digitalWrite(13, LOW);
    //以下代码在开机时等待MPU6050进入稳定状态
    //需要手扶小车,定位在平衡位置5度范围内
    //直到传感器稳定才会进入平衡程序
    //如果6秒钟内无法达到平衡,看门狗会复位芯片,并重复此过程
    //轻微地摇晃小车能加速稳定过程
    while (!mpuToken)
        ;
    mpuToken = 0;
    float angle_avg = angle;
    while (1) {
        if (!mpuToken) continue;
        mpuToken = 0;
        if (millis() < 5000) wdt_reset();
        angle_avg = angle_avg * 0.9 + angle + 0.1;
        Serial.println(angle);
        if (abs(angle_avg) < 5) break;
    }
}

/*
 * 主逻辑
 * 流程1:串口接收遥控数据,更新姿态变量
 * 流程2:根据姿态数据,设置当前的角度误差,速度误差
 * 流程3:根据控制需要,设置角度误差/速度误差的偏置值,达到遥控目的
 * 流程4:使用pid计算左轮右轮速度
 * 流程5:对计算后的速度进行限幅,映射等后处理.
 * 流程6:设置马达PWM
 * 流程7:超声波检测,保持以上步骤为固定周期调用
 */
float ultrPing = 350;
float angle_error_detect = 0;
unsigned char oldToken = 0;
void loop() {
    digitalWrite(13, !digitalRead(13));
    if (oldToken == mpuToken) return;  //约10ms有一次数据
    oldToken = mpuToken;
    // MPU6050会因为电压不稳定而崩溃,异常波动会对马达驱动造成伤害
    //此代码检测巨大的角度变化(正常情况下不会发生)
    //发生时将停止马达驱动,并等待看门狗复位芯片
    while (abs(angle - angle_error_detect) > 150) {
        analogWrite(3, 0);
        analogWrite(11, 0);
    }
    angle_error_detect = angle;
    output();
    wdt_reset();
    //超声波检测
    if (digitalRead(2)) return;  //超声波忙
    pinMode(2, OUTPUT);
    digitalWrite(2, HIGH);  //给测量信号
    delayMicroseconds(10);
    digitalWrite(2, LOW);  //输入模式,等待信号
    pinMode(2, INPUT);
    //探测时间限制2000us,即2毫秒,对应的最大探测距离约为350毫米
    int trigTime = pulseIn(2, HIGH, 2000);
    float ping = trigTime / 5.8;
    if (trigTime == 0) ping = 2000.0 / 5.8;
    ultrPing = ultrPing * 0.7 + ping * 0.3;
}

/*
 * 区分左右
 * 人与小车同向(前进方向),此时左手边为左马达,由speedL驱动,speedL为正数时,轮子驱动小车前进,右边同理
 * 小车向前(前进方向)倾斜时,角度传感器输出正角度,反之输出负角度
 * 马达驱动小车前进时(PWM为正数),对应编码器输出正值,后退时输出负值
 *
 * 接线
 * 超声波模块接数字引脚2
 *
 * 左边马达接马达驱动M1接口,由数字引脚11输出pwm,数字引脚12输出方向信号
 * 右边马达接马达驱动M2接口,由数字引脚3输出PWM,数字引脚8输出方向信号
 * 检查确认马达的左右顺序,前进后退方向
 *
 * 左边马达对应的编码器A相接模拟引脚A0, B相接模拟引脚A1
 * 右边马达对应的编码器A相接模拟引脚A2, B相接模拟引脚A3
 * 检查马达PWM为正时,编码均输出正值
 *
 * 姿态传感器MPU6050
 * SCL引脚接模拟引脚A5,SDA引脚接模拟引脚A4
 * 检查,当小车向前进方向倾斜时,输出值为正数,反之输出负数
 *
 */

/*
 * 注意事项:
 *
 * 马达没有转动时,会有电流声,属于正常情况,修改PWM的频率可以去掉电流声,但是不要这么做.
 * 平衡车在大部分时间都是静止的,在每个PWM周期,供电的瞬间都有5us时间近似短路状态
 * Arduino的默认频率为1khz,意味着1秒钟有5ms近似短路,如果将频率修改为40khz,就有40ms时间近似短路
 * 此举将很容易导致马达驱动损毁
 *
 * 此小车配置的马达为高响应马达,意味着内阻低,堵转电流为8安培,为通常马达的4倍.
 * 通过示波器观察,近似短路期间,会将电池电压拉低2~3V.
 * 如果电池电量不足,或者电池内阻大,可能会导致芯片复位
 * 尤其是MPU6050,受电压影响时,将不能正确解算姿态,从串口监视器看到的值几乎为乱码,马达表现为乱颤
 * 此时也比较容易损坏马达驱动
 * 马达驱动模块配置了1.1安培的自恢复保险丝,可以在一定程度上保护了驱动模块.
 * 副作用是,频繁换向时,有可能会触发保险丝熔断,表现为马达只能非常低速旋转.
 * 停机等待片刻即可恢复
 * 马达驱动模块损坏通常是有一个方向可以正常速度,另一方向异常
 *
 * 小车开机后,会等待姿态传感器MPU6050稳定,才会开始平衡控制,
 * 需要手持小车,轮子不要悬空,保持小车在平衡位置附近(5度范围内)
 * 轻微摇晃可以加速稳定过程,直到马达有输出
 */

/*
 * helper_3dmath.h,I2Cdev.h,I2Cdev.cpp,MPU6050.h,MPU6050.cpp
 * 均来自开源库: https://github.com/jrowberg/i2cdevlib
 *
 * enc_2wbt.ino
 * 主文件,初始化,读取姿态,超声波检测
 *
 * calcAngle.ino
 * 读取MPU6050 fifo,解算出当前倾斜角度,需要设置机械原点变量offset_angle
 *
 * encoder.ino
 * 编码器中断处理函数,解码编码器,并记录编码器绝对值
 *
 * remote.ino
 * 处理从串口接收到得遥控信号,赋值给rc_forward,rc_truen变量
 *
 * pidOutput.ino
 * 从其他文件引入需要得变量(倾角,编码器,遥控器),计算pwm并输出
 *
 * check.ino
 * 检查组装/接线是否正确
 */